#include <Magick++.h>
#include <iostream>
#include <sstream>
#include <vector>
#include <cmath>

#include <getopt.h>

#include "chrono.h"

//#define ONLY_BILATERAL
#define POW2(a) ((a) * (a))

using namespace std;
using namespace Magick;

typedef unsigned int size_type;
typedef double real_type;

// Resize image
void resize( 
	Image &image, 
	const Geometry &geometry_,
	const real_type blur=0.92, 
	MagickCore::FilterTypes filter=LanczosFilter )
{
  // Calculate new size.  This code should be supported using binary arguments
  // in the ImageMagick library.
  long x = 0;
  long y = 0;
  unsigned long width = image.columns();
  unsigned long height = image.rows();

	image.filterType(filter);

	MagickCore::ParseMetaGeometry (static_cast<std::string>(geometry_).c_str(),
                     &x, &y,
                     &width, &height );

	MagickCore::ExceptionInfo exceptionInfo;
	MagickCore::GetExceptionInfo( &exceptionInfo );
	MagickCore::Image* newImage =
    ResizeImage( image.image(),
               width,
               height,
               image.image()->filter,
               blur,
               &exceptionInfo);
	image.replaceImage( newImage );
	throwException( exceptionInfo );
	(void) MagickCore::DestroyExceptionInfo( &exceptionInfo );
}


inline int image_sample(int pos, const int &max)
{
	if(pos < 0)
		pos = -pos;
	else if(pos > max)
		pos = max - (pos - max);

	return pos;
}

MagickCore::FilterTypes getFilterType(unsigned int filter)
{

	MagickCore::FilterTypes filterType;
	switch(filter)
	{
		case 1: 
			filterType = (PointFilter);
			break;
		case 2: 
			filterType = (BoxFilter);
			break;
		case 3: 
			filterType = (TriangleFilter);
			break;
		case 4: 
			filterType = (HermiteFilter);
			break;
		case 5: 
			filterType = (HanningFilter);
			break;
		case 6: 
			filterType = (HammingFilter);
			break;
		case 7: 
			filterType = (BlackmanFilter);
			break;
		case 8: 
			filterType = (GaussianFilter);
			break;
		case 9: 
			filterType = (QuadraticFilter);
			break;
		case 10: 
			filterType = (CubicFilter);
			break;
		case 11: 
			filterType = (CatromFilter);
			break;
		case 12: 
			filterType = (MitchellFilter);
			break;
		case 13:
			filterType = (LanczosFilter);
			break;
		case 14:
			filterType = (BesselFilter);
			break;
		case 15:
			filterType = (SincFilter);
			break;
		default:
			filterType = (UndefinedFilter);
	}

	return filterType;

}

void downsampling(
	Image &img, 
	real_type downsample_factor=4.0, 
	real_type blur=1.0, 
	size_type filter=10)
{
    const size_type width = img.columns();
    const size_type height= img.rows();
    const size_type small_width  = static_cast<size_type>(width / downsample_factor);
    const size_type small_height = static_cast<size_type>(height / downsample_factor);

	cout << "Info" << endl;
	cout << "\tInput resolution: " << width << "," << height << endl;
	cout << "\tDownsampling resolution: " << small_width << "," << small_height << endl;
	cout << "\tScales: " << 1.0*small_width/width << "," << 1.0*small_height/height << endl;

	cout << endl << "Downsampling image " << img.fileName() << endl;
#ifdef CHRONO
    Chrono chrono_down("downsampling");
    chrono_down.start();
#endif


	Image small_img = img;
	small_img.modifyImage();
	//small_img.scale(Geometry(small_width,small_height));

	//Tip: Garantindo que o resize utilize exatamente o tamnho original da imagem.
	stringstream ss;
	ss << small_width << "x" << small_height << "!";
	resize(small_img,Geometry(ss.str()), blur, getFilterType(filter));

    small_img.syncPixels();
    small_img.write("downsampled_image.png");
    //small_img.write("downsampled_image.ppm");

#ifdef CHRONO
  chrono_down.stop();
  cout<<"  "<<chrono_down.report()<<endl;
#endif

	cout << "The image was saved as downsampled_image.png at current directory." << endl;

}

void joint_bilateral_upsampling(
	Image &img,
	Image &small_img,
	double downsample_factor=4.0,
	double blur=0.92, size_type filter=10,
	double space_radius=2.0,
	double range_radius=2.0,
	double space_sigma=6.0,
	double range_sigma=.1)
{

    const size_type width = img.columns();
    const size_type height= img.rows();

    size_type small_width  = static_cast<size_type>(width / downsample_factor);
    size_type small_height = static_cast<size_type>(height / downsample_factor);

    PixelPacket *input = img.getPixels(0,0,width,height);

#ifndef ONLY_BILATERAL
	if(small_img.columns() != small_width || small_img.rows() != small_height)
	{
		fprintf(stderr,"Warning:\n\tThe downsampled image should have this resolution: (%d,%d)\n\n", small_width,small_height); 
		small_width = small_img.columns();
		small_height =  small_img.rows();
		//FIXME: Pra usar as imagens do Cicconet, por hora vai passar assim mesmo!
		//return;
	}
#endif

    Image img_output(Geometry(width,height), "black");
    PixelPacket *result = img_output.getPixels(0,0,width,height);

    img.modifyImage();
    small_img.modifyImage();
    img_output.modifyImage();


	cout << "Info" << endl;
	cout << "\tInput resolution: " << width << "," << height << endl;
	cout << "\tDownsample resolution: " << small_width << "," << small_height << endl;
	cout << "\tScales: " << 1.0*small_width/width << "," << 1.0*small_height/height << endl;
	cout << "\tDownsample factor: " << downsample_factor  << endl;
	cout << "\tReal downsample factor: " << 1.0*width/small_width << "," << 1.0*height/small_height << endl;
	cout << "\tRange sigma:  " << range_sigma << endl;
	cout << "\tRange radius: " << range_radius <<  " ( "<< (2*range_radius)+1 << "x" <<  (2*range_radius)+1 << " )" << endl;
	cout << "\tSpace sigma:  " << space_sigma << endl;
	cout << "\tSpace radius: " << space_radius <<  " ( "<< (2*space_radius)+1 << "x" <<  (2*space_radius)+1 << " )" << endl; 
	cout << "\tUpsample blur:" << blur << endl;
	cout << "\tFilter type:  " << filter << endl;
	cout << endl;

#ifdef CHRONO
    Chrono chrono("filter");
    chrono.start();
#endif


#ifndef ONLY_BILATERAL


#ifdef CHRONO
    Chrono chrono_up("upsample");
    chrono_up.start();
#endif

	cout << endl << "Pre-upsample solution" << endl;

	Image up_img(small_img);
	up_img.modifyImage();


	//Tip: Garantindo que o resize utilize exatamente o tamnho original da imagem.
	stringstream ss;
	ss << width << "x" << height << "!";
	resize( up_img, Geometry(ss.str()), blur, getFilterType(filter));

    PixelPacket *up_input = up_img.getPixels(0,0,width,height);


#ifdef CHRONO
	chrono_up.stop();
	cout<<"  "<<chrono_up.report()<<endl;
#endif


#endif //ONLY_BILATERAL


#ifdef CHRONO
    Chrono chrono_gauss("gaussian");
    chrono_gauss.start();
#endif


	cout << endl << "Space domain gaussian..." << endl;

	int iradius = space_radius;

	vector<real_type> gauss;
	size_type gauss_size = iradius*2 + 1;
	gauss.reserve(POW2(gauss_size));

	real_type spacial_gaussian_height = ( 1.0 / (2.0*M_PI*POW2(space_sigma)) );

	for(int y=-iradius; y <= iradius; ++y)
	{
		for(int x=-iradius; x <= iradius; ++x)
		{

			const real_type rr = ( POW2(x) + POW2(y) ) / POW2(space_sigma);
			gauss[(x+iradius) + ( (y+iradius) * gauss_size )] = (spacial_gaussian_height * exp(-rr*0.5));

			printf("%.6f   \t", gauss[x+iradius + ((y+iradius)*gauss_size)]);

		}
		cout << endl;
	}
	//return;


#ifdef CHRONO
  chrono_gauss.stop();
  cout<<"  "<<chrono_gauss.report()<<endl;

  Chrono chrono_jbu("JBU");
  chrono_jbu.start();
#endif


#ifdef ONLY_BILATERAL
	cout << endl << "Bilateral..." << endl;
#else
	cout << endl << "Joint Bilateral Upsampler..." << endl;
#endif


	//bool PRINT=false;

#ifdef ONLY_BILATERAL
	const int k = static_cast<size_type>(space_radius - range_radius) + iradius;
#endif

	iradius = range_radius;
	const real_type scale = 1.0 / downsample_factor;
	const real_type square_range_sigma = 1.0 / POW2(range_sigma);
	const real_type range_gaussian_height = ( 1.0 / ( 2.0*M_PI*POW2(range_sigma) ) ); //2D gaussian
	//const real_type range_gaussian_height = ( 1.0 / ( sqrt(2.0*M_PI)*(range_sigma) ) ); //1D gaussian

    for(size_type y=0; y < height; ++y)
	{
		for(size_type x=0; x < width; x++)
		{

			PixelPacket *p = input + (x + (y*width));
			real_type p_r = Color::scaleQuantumToDouble(p->red);
			real_type p_g = Color::scaleQuantumToDouble(p->green);
			real_type p_b = Color::scaleQuantumToDouble(p->blue);

			real_type accumulated[4] = {0.,0.,0.,0.};
			real_type Wp = 0.0;

			for(int v=-iradius; v <= iradius; v++)
			{
				for(int u=-iradius; u <= iradius; u++)
				{
					int i = x + u;
					int j = y + v;

					int pos = (image_sample(i,width-1) + image_sample(j,height-1)*width);

					//Tip:	Considerando somente os pixels q tal que || p - q || <= 3*space_sigma.
					//		A razão é que os pixels mais afastados que 2*space_sigma influenciam pouco no resultado.
					//		A complexidade fica O( N*range_radius^2 )
					//Nota:
					//		Apesar de ficar mais rápido, em imagens grandes algumas arestas não são preservadas, é o que parece!
#if 1
					const int pq = (x + (y*width)) - pos;
					//if( pq > 3*space_sigma) break;
					if( pq > 3*space_sigma) continue;
#endif

					PixelPacket *q = input + pos;
					real_type q_r = Color::scaleQuantumToDouble(q->red);
					real_type q_g = Color::scaleQuantumToDouble(q->green);
					real_type q_b = Color::scaleQuantumToDouble(q->blue);

#ifdef ONLY_BILATERAL
					real_type gaussian_weight = gauss[ (u+k) + ( (v+k) * gauss_size )];
					//Usa a cor da imagem original, é simplesmente um filtro bilateral.
					PixelPacket *q_small = q;
#else
					//FIXME: otimizar!!!
					const real_type gaussian_weight = gauss[ 
							( static_cast<size_type>((x + i ) * scale) % gauss_size ) + 
							( static_cast<size_type>((y + j ) * scale) % gauss_size  )*gauss_size
						];

					PixelPacket *q_small = up_input + pos;
#endif //ONLY_BILATERAL

					real_type rr = (POW2(p_r - q_r) + POW2(p_g - q_g) + POW2(p_b - q_b)) * square_range_sigma;
					real_type diff_map = range_gaussian_height * exp(-rr*0.5);

					real_type weight = gaussian_weight * diff_map;

					real_type q_r_small = Color::scaleQuantumToDouble(q_small->red);
					real_type q_g_small = Color::scaleQuantumToDouble(q_small->green);
					real_type q_b_small = Color::scaleQuantumToDouble(q_small->blue);

					accumulated[0] += q_r_small * weight;
					accumulated[1] += q_g_small * weight;
					accumulated[2] += q_b_small * weight;

					Wp += weight;
				}

			}


			accumulated[0] /= Wp;
			accumulated[1] /= Wp;
			accumulated[2] /= Wp;

			//ColorRGB color(accumulated[0],accumulated[1],accumulated[2]);
			//*result++ = color;
			result->red	  = Color::scaleDoubleToQuantum(accumulated[0]);
			result->green = Color::scaleDoubleToQuantum(accumulated[1]);
			result->blue  = Color::scaleDoubleToQuantum(accumulated[2]);
			result++;
		}
	}


#ifdef CHRONO
	chrono_jbu.stop();
	cout<<"  "<<chrono_jbu.report()<<endl;

	Chrono chrono_write("writing");
	chrono_write.start();
#endif

    cout << endl << "Writing result" << endl;

#ifndef ONLY_BILATERAL
    up_img.syncPixels();
	//up_img.write("UP.png");
#endif

    img.syncPixels();
    small_img.syncPixels();
    img_output.syncPixels();
    img_output.write("result_image.png");

#ifdef CHRONO
	chrono_write.stop();
	cout<<"  "<<chrono_write.report()<<endl;

	chrono.stop();
	cout<<"  "<<chrono.report()<<endl;
#endif
}

void print_help(const char *program)
{
	cout 
		<<  "Usage:" << endl
		<<  "  " << program << " -f <filename> -F <downsampled_filename> [OPTION] [FILTER]" << endl
		<< endl
		<<  "Usage only for downsample: " << endl
		<<  "  " << program << " -f <filename>" << endl
		<< endl << endl
		<<  "Options:" << endl
		<<  "  -s, --space-radius\t\tRadius used by space gaussian." << endl
		<<  "  -r, --range-radius\t\tRadius used by range gaussian." << endl
		<<  "  -S, --space-sigma\t\tSigma used by space gaussian." << endl
		<<  "  -R, --range-sigma\t\tSigma used by range gaussian." << endl
		<<  "  -d, --downsample-factor\tDownsample factor used." << endl
		<<  "  -f, --filename\t\tImage filename." << endl
		<<  "  -F, --downsample-filename\tDownsampled image filename." << endl
		<<  "  -b, --resize-blur\t\tReduce the image noise on resize." << endl
		<<  "  -t, --resize-filter\t\tFilter type used on resize." << endl
		<<  "  -h, --help\t\t\tPrint this message." << endl
		<< endl
		<<  "Filters:" << endl
		<<  "  0:\tUndefinedFilter" << endl
		<<  "  1:\tPointFilter" << endl
		<<  "  2:\tBoxFilter" << endl
		<<  "  3:\tTriangleFilter" << endl
		<<  "  4:\tHermiteFilter" << endl
		<<  "  5:\tHanningFilter" << endl
		<<  "  6:\tHammingFilter" << endl
		<<  "  7:\tBlackmanFilter (default)" << endl
		<<  "  8:\tGaussianFilter" << endl
		<<  "  9:\tQuadraticFilter" << endl
		<<  "  10:\tCubicFilter" << endl
		<<  "  11:\tCatromFilter" << endl
		<<  "  12:\tMitchellFilter" << endl
		<<  "  13:\tLanczosFilter" << endl
		<<  "  14:\tBesselFilter" << endl
		<<  "  15:\tSincFilter" << endl
	<< endl;
}

int main(const int argc, char **argv)
{

	int c;
	static int verbose_flag;
	static int downsample_flag;
	unsigned int filter_type=7;
	double	space_radius=2.0, space_sigma=4.0,
			range_radius=2.0, range_sigma=0.05,
			downsample_factor=4.0, blur=0.92;
	string filename, down_filename;

	while(1)
	{
		static struct option long_options[] = 
		{
			// These options set a flag.
			{"verbose",		no_argument,	&verbose_flag,		1},
			{"brief",		no_argument,	&verbose_flag,		0},
			{"downsample",	no_argument,	&downsample_flag,	1},
			{"no-downsample", no_argument,	&downsample_flag,	0},
			/*
			These options don't set a flag.
			We distinguish them by their indices.
			*/
			{"space-radius",		required_argument,	0,	's'},
			{"range-radius",		required_argument,	0,	'r'},
			{"space-sigma",			required_argument,	0,	'S'},
			{"range-sigma",			required_argument,	0,	'R'},
			{"downsample-factor",	required_argument,	0,	'd'},
			{"filename",			required_argument,	0,	'f'},
			{"downsample-filename",	required_argument,	0,	'F'},
			{"resize-blur",			required_argument,	0,	'b'},
			{"resize-filter",		required_argument,	0,	't'},
			{"help",				no_argument,		0,	'?'},
			{0, 0, 0, 0}
		};

		/* getopt_long stores the option index here */
		int option_index = 0;

		c = getopt_long(argc, argv, "s:r:S:R:d:f:F:b:t:?h", long_options, &option_index);

		/* Detect the end of the options. */
		if(c == -1)
		{
			break;
		}

		stringstream ss(stringstream::in | stringstream::out);
		ss << optarg;

		switch(c)
		{
			case 0:
				if(long_options[option_index].flag != 0)
					break;
				printf("option %s", long_options[option_index].name);
				if(optarg)
					printf("with arg %s", optarg);
				printf("\n");
				break;

			case 's':
				ss >> space_radius;
				break;
			case 'r':
				ss >> range_radius;
				break;
			case 'S':
				ss >> space_sigma;
				break;
			case 'R':
				ss >> range_sigma;
				break;
			case 'f':
				ss >> filename;
				break;
			case 'F':
				ss >> down_filename;
				break;
			case 'd':
				ss >> downsample_factor;
				break;
			case 'b':
				ss >> blur;
				break;
			case 't':
				ss >> filter_type;
				break;
			case '?':
			case 'h':
				print_help(argv[0]);
				return 0;
			default:
				abort();
		}
	}

	if(optind < argc)
	{
		printf("non-option ARGV-elements: ");
		while(optind < argc)
			printf("%s ", argv[optind++]);
		putchar('\n');

		return 1;
	}


    try
    {
		if(filename.empty())
		{
			print_help(argv[0]);
			return 1;
		}


		Image input;
        input.read(filename);

		if(downsample_flag || down_filename.empty())
		{
			printf("Downsample image using downsample factor %f\n", downsample_factor);
			downsampling(input, downsample_factor, blur, filter_type);
			return 0;
		}

		if(down_filename.empty())
		{
			cout << "Remenber you should inform the filename of the downsampled image" << endl;
			print_help(argv[0]);
			return 1;
		}

		Image small_input;
		small_input.read(down_filename);

      joint_bilateral_upsampling(
			input, small_input,
			downsample_factor, 
			blur, filter_type,
			space_radius, range_radius,
			space_sigma, range_sigma);
    }
    catch(Exception &error)
    {
        cout << "Caught exception: " << error.what() << endl;
        return 1;
    }

    return 0;
}
